//////////////////////////////////////////////////////////////
//	: tet.h												//
//	           	//
//	Copyright (c) 2010-2013   	//
//////////////////////////////////////////////////////////////

#ifndef _TEMPLET_RUN_TIME
#define _TEMPLET_RUN_TIME

#include <vector>

namespace TEMPLET {

class Assemble;
class Channel;
class Process;
class Activator;

class Assemble{
	friend class Channel; 
	friend class Activator;
protected:
	virtual ~Assemble();
	Assemble(int NT=0){}
public:	
	bool run();
protected:
	void _regChan(Channel*chan){channels.push_back(chan);}
	void _regProc(Process*prc){processes.push_back(prc);}
private:
	void _send(Channel*c){ready.push_back(c);}
private:
	std::vector<Channel*> channels;
	std::vector<Process*> processes;
	std::vector<Channel*> ready;
};

class Channel{
	friend class Assemble;
	friend class Process;
	friend class Activator;
protected:
	Channel(Assemble*a):_assemble(a),_active(CLI),
		_cliPort(0),_srvPort(0),_cli_selector(0),_srv_selector(0){}
	virtual ~Channel(){}
public:
	enum ChanActivity{CLI,SRV,RTL_CLI,RTL_SRV};
protected:	
	void _send(){_assemble->_send(this);}
	ChanActivity _active;
public:
	void resend();
private:
	bool _run();
public:
	Process* _cliPort;
	Process* _srvPort;
	int _cli_selector;
	int _srv_selector;
private:
	Assemble* _assemble;
};

class Activator:public Channel{
public:
	Activator(Assemble*a,Process*p):Channel(a){_cliPort=p;}
	void _send(int port){_cli_selector=port;_active=RTL_CLI;_assemble->_send(this);}
};

class Process{
	friend class Channel;
	friend class Assemble;
protected:
	void lock(){}
	void unlock(){}
protected:
	Process(Assemble*a):_assemble(a){}
	virtual ~Process(){}
protected:
	Activator* _createActivator(){return new Activator(_assemble,this);}
private:
	virtual void _run(int _selector,Channel*)=0;
private:
	Assemble* _assemble;
};

}
#endif