//////////////////////////////////////////////////////////////
//	: tbag.cpp											//
//	 ' '  POSIX pthread		      	//
//	Copyright (c) 2010-2013   	//
//////////////////////////////////////////////////////////////

#include "tbag.h"
#include <assert.h>

namespace Taskbag {

TBag::TBag(int num_prc,int argc, char* argv[])
{
	nproc=num_prc;
	task=new Task*[nproc];
	thread=new pthread_t[nproc];
	pthread_mutex_init(&cs,NULL);
}

TBag::~TBag()
{
	delete task;
	delete thread;
	pthread_mutex_destroy(&cs);
}
	
void TBag::run()
{
	cur_task=0;
	c_active=0;

	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);

	for(int i=0;i<nproc;i++){
		int res=pthread_create(&thread[i],&attr,tFunc,this);
		task[i]=createTask();
		assert(res==0 && task[i]);
	}
	pthread_cond_init(&await,NULL);
	//run
	for(int i=0;i<nproc;i++){
		pthread_join(thread[i], NULL);
		delete task[i];
	}
	pthread_attr_destroy(&attr);
	pthread_cond_destroy(&await);
}

void* tFunc(void* p)
{
	TBag* b=(TBag*)p;
	TBag::Task* task;

	pthread_mutex_lock(&b->cs);
	task=b->task[b->cur_task++];
	for(;;){
		while(!b->if_job()){
			if(!b->c_active){
				pthread_cond_signal(&b->await);
				pthread_mutex_unlock(&b->cs);
				return 0;
			}
			pthread_cond_wait(&b->await,&b->cs);
		}
		b->get(task);
		b->c_active++;
		pthread_mutex_unlock(&b->cs);

		b->proc(task);
		
		pthread_mutex_lock(&b->cs);
		b->c_active--;
		b->put(task);
		pthread_cond_signal(&b->await);
	}
	return 0;
}
}
